Require Export PropiLmodel.

(** * Independence of Instuitionistic Connectives *)

(** Disjunction property *)

Fixpoint HarropF u: Prop :=
  match u with
    | Var _ => True
    | Fal => True
    | Imp s t => HarropF t
    | And s t => HarropF s /\ HarropF t
    | Or s t => False
  end.

Definition HarropC A: Prop := forall s, s el A -> HarropF s.

Lemma Harrop A p q: HarropC A -> (gen A (Or p q) <-> gen A p \/ gen A q).
Proof.
  intros C. split.
  - intros D. remember (Or p q) as B.
    induction D as [| | |A s t v E F _ G IHG|A s t u E F IH| |
                  A s t u E _ _ _ _|A s t E _|A s t E _];
    try congruence; subst; auto.
    + assert (H: HarropC (t::A)).
        { apply C in E. intros u [I|I]; subst; auto. }
      assert (K: gen (t :: A) p \/ gen (t :: A) q) by (apply IHG; auto).
      destruct K as [K|K].
      * left. apply genIL with (s:=s) (t:=t); auto.
      * right. apply genIL with (s:=s) (t:=t); auto.
    + assert (H: HarropC (s::t::A)).
        { apply C in E as [E1 E2]. intros u [I|[I|I]]; subst; auto. }
      assert (K: gen (s :: t :: A) p \/ gen (s :: t :: A) q) by (apply IH; auto).
      destruct K as [K|K].
      * left. apply genAL with (s:=s) (t:=t); auto.
      * right. apply genAL with (s:=s) (t:=t); auto.
    + exfalso. apply (C (Or s t) E).
    + assert (s=p) by congruence. subst. auto.
    + assert (t=q) by congruence. subst. auto.
  - intros [D|D]; auto.
Qed.

Lemma OrFree_Harrop s: OrFree s -> HarropF s.
Proof. induction s; simpl; auto; intros [D E]; auto. Qed.

Lemma genXM x:
  ~ gen [] (Or (Var x) (Not (Var x))).
Proof.
  intros C. apply Harrop in C.
  - destruct C as [C|C]; revert C.
    + apply gen_NoVar.
    + apply gen_NoNVar.
  - intros s F. exfalso. apply F.
Qed.


Lemma OrIndependence p q:
   ~ (gen [p] q \/ gen [q] p)
   -> ~ exists s, OrFree s /\ gen [] (Equi (Or p q) s).
Proof.
  intros F [s [C D]]. unfold Equi in D.
  destruct (gen_and [] (Imp (Or p q) s) (Imp s (Or p q))) as [E _].
  specialize (E D).
  assert (H: gen [s] (Or p q)).
  { apply gen_cut with (s:=And (Imp p (Or p q)) (Imp q (Or p q))); auto.
    remember (Or p q) as r.
    assert (H: gen [] (Imp s r)) by auto.
    apply gen_imp in H. apply (genW H); auto. }
  assert (K: gen [s] p \/ gen [s] q).
  { apply Harrop; auto. intros u [K|K]; subst; auto using OrFree_Harrop.
    exfalso. auto. }
  assert (I1: gen [p] s).
  { assert (I: gen [Or p q] s) by (apply gen_imp; auto).
    apply gen_cut with (s:=Or p q); eauto using genW. }
  assert (I2: gen [q] s).
  { assert (I: gen [Or p q] s) by (apply gen_imp; auto).
    apply gen_cut with (s:=Or p q); eauto using genW. }
  apply F. destruct K as [K|K].
  - right. apply gen_cut with (s:=s); eauto using genW.
  - left. apply gen_cut with (s:=s); eauto using genW.
Qed.



(** * Undefinability due to McKinsey *)

(* The results are from
  McKinsey, J. C. C. (1939). Proof of the independence of the primitive symbols
  of Heyting's calculus of propositions. Journal of Symbolic Logic 4 (4):155-158
*)

(** * Independence of Falsehood *)
Section HeytingHMK1.
  (* McKinsey's Matrix I to prove undefinability of Fal *)

  (* Lattice:
      top
       |
       |
       a
       |
       |
      bot
  *)

  Inductive HMK1: Type := HMK1bot | HMK1a | HMK1top.

  Definition HMK1R (a b: HMK1) : Prop :=
    match a,b with
      | HMK1bot, _ => True
      | _, HMK1bot => False
      | _, HMK1top => True
      | HMK1top, _ => False
      | HMK1a, HMK1a => True
    end.

  Definition HMK1meet (a b: HMK1) : HMK1 :=
    match a,b with
      | HMK1bot, _ => HMK1bot
      | _ , HMK1bot => HMK1bot
      | HMK1top, b => b
      | a, HMK1top => a
      | HMK1a, HMK1a => HMK1a
    end.

  Definition HMK1join (a b: HMK1) : HMK1 :=
    match a,b with
      | HMK1top, _ => HMK1top
      | _, HMK1top => HMK1top
      | HMK1bot, b => b
      | a, HMK1bot => a
      | HMK1a, HMK1a => HMK1a
    end.

  Definition HMK1imp (a b : HMK1) : HMK1 :=
    match a,b with
      | HMK1bot, _ => HMK1top
      | _, HMK1top => HMK1top
      | _, HMK1bot => HMK1bot
      | HMK1top, b => b
      | HMK1a, HMK1a => HMK1top
    end.

  Lemma HMK1Rref a: HMK1R a a.
  Proof. destruct a; simpl; tauto. Qed.

  Lemma HMK1Rtra a b c: HMK1R a b -> HMK1R b c -> HMK1R a c.
  Proof. destruct a,b,c; simpl; tauto. Qed.

  Definition HMK1HA : HeytingAlgebra.
  Proof.
    apply (@mkHeytingAlgebra HMK1 HMK1R HMK1bot HMK1meet HMK1join HMK1imp).
    - exact HMK1Rref.
    - exact HMK1Rtra.
    - intros u. destruct u; simpl; tauto.
    - intros u v w. destruct u, v, w; simpl; tauto.
    - intros u v w. destruct u, v, w; simpl; tauto.
    - intros u v w. destruct u, v, w; simpl; tauto.
  Defined.

  (* Fal is undefinable by other connectives *)
  Lemma FalIndependence_hey:
    ~ exists s, FalFree s /\ nd [] (Equi s Fal).
  Proof.
    intros [s [C D]].
    pose (HMK1interp (_ : var) := HMK1top).
    assert (H: @evalH HMK1HA HMK1interp s = HMK1top).
    { clear D. induction s; simpl; auto.
      - exfalso. exact C.
      - destruct C as [C1 C2]. rewrite IHs1; auto. rewrite IHs2; auto.
      - destruct C as [C1 C2]. rewrite IHs1; auto. rewrite IHs2; auto.
      - destruct C as [C1 C2]. rewrite IHs1; auto. }
    apply (@HAequiv HMK1HA HMK1interp) in D as [D _]. rewrite H in D.
    exact D.
  Qed.

End HeytingHMK1.

(** * Independence of Disjunction *)
Section HeytingHMK2.
  (* McKinsey's Matrix II to prove undefinability of Or *)

  (* Lattice: 
                        top
                         |
                         |
                      a \/ b
                     /      \
                    a        b
                     \      /
                      \    /
                       bot
  *)

  Inductive HMK2 : Type := HMK2bot | HMK2a | HMK2b | HMK2ab | HMK2top.

  Definition HMK2R (a b: HMK2) : Prop :=
    match a,b with
      | HMK2bot, _ => True
      | _, HMK2bot => False
      | _, HMK2top => True
      | HMK2top, _ => False
      | HMK2a, HMK2a => True
      | HMK2a, HMK2b => False
      | HMK2a, HMK2ab => True
      | HMK2b, HMK2a => False
      | HMK2b, HMK2b => True
      | HMK2b, HMK2ab => True
      | HMK2ab, HMK2a => False
      | HMK2ab, HMK2b => False
      | HMK2ab, HMK2ab => True
    end.

  Definition HMK2meet (a b: HMK2) : HMK2 :=
    match a,b with
      | HMK2bot, _ => HMK2bot
      | _, HMK2bot => HMK2bot
      | HMK2top, b => b
      | a, HMK2top => a
      | HMK2a, HMK2a => HMK2a
      | HMK2a, HMK2b => HMK2bot
      | HMK2a, HMK2ab => HMK2a
      | HMK2b, HMK2a => HMK2bot
      | HMK2b, HMK2b => HMK2b
      | HMK2b, HMK2ab => HMK2b
      | HMK2ab, b => b
    end.

  Definition HMK2join (a b: HMK2) : HMK2 :=
    match a,b with
      | HMK2bot, b => b
      | a, HMK2bot => a
      | HMK2top, _ => HMK2top
      | _, HMK2top => HMK2top
      | _, HMK2ab => HMK2ab
      | HMK2ab, _ => HMK2ab
      | HMK2a, HMK2a => HMK2a
      | HMK2a, HMK2b => HMK2ab
      | HMK2b, HMK2a => HMK2ab
      | HMK2b, HMK2b => HMK2b
    end.

  Definition HMK2imp (a b: HMK2) : HMK2 :=
    match a,b with
      | HMK2bot, _ => HMK2top
      | _, HMK2top => HMK2top
      | HMK2a, HMK2bot => HMK2b
      | HMK2a, HMK2a => HMK2top
      | HMK2a, HMK2b => HMK2b
      | HMK2a, HMK2ab => HMK2top
      | HMK2b, HMK2bot => HMK2a
      | HMK2b, HMK2a => HMK2a
      | HMK2b, HMK2b => HMK2top
      | HMK2b, HMK2ab => HMK2top
      | HMK2ab, HMK2bot => HMK2bot
      | HMK2ab, HMK2a => HMK2a
      | HMK2ab, HMK2b => HMK2b
      | HMK2ab, HMK2ab => HMK2top
      | HMK2top, b => b
    end.

  Lemma HMK2Rref a: HMK2R a a.
  Proof. destruct a; simpl; tauto. Qed.

  Lemma HMK2Rtra a b c: HMK2R a b -> HMK2R b c -> HMK2R a c.
  Proof. destruct a,b,c; simpl; tauto. Qed.

  Definition HMK2HA : HeytingAlgebra.
  Proof.
  apply (@mkHeytingAlgebra HMK2 HMK2R HMK2bot HMK2meet HMK2join HMK2imp).
  - exact HMK2Rref.
  - exact HMK2Rtra.
  - intros u. destruct u; simpl; tauto.
  - intros u v w. destruct u, v, w; simpl; tauto.
  - intros u v w. destruct u, v, w; simpl; tauto.
  - intros u v w. destruct u, v, w; simpl; tauto.
  Defined.

  Lemma OrIndependence_hey:
    ~ exists s, OrFree s /\ nd [] (Equi (Or (Var 0) (Var 1)) s).
  Proof.
    intros [s [C D]].
    pose (HMK2interp (z: var) := match z with | 0 => HMK2a | _ => HMK2b end).
    assert (H: @evalH HMK2HA HMK2interp s <> HMK2ab).
    { clear D. induction s; simpl; auto; try congruence.
      - destruct v; simpl; congruence.
      - destruct C as [C1 C2]. specialize (IHs1 C1). specialize (IHs2 C2).
        destruct (evalH (HA:=HMK2HA) HMK2interp s1);
        destruct (evalH (HA:=HMK2HA) HMK2interp s2); simpl; congruence.
      - destruct C as [C1 C2]. specialize (IHs1 C1). specialize (IHs2 C2).
        destruct (evalH (HA:=HMK2HA) HMK2interp s1);
        destruct (evalH (HA:=HMK2HA) HMK2interp s2); simpl; congruence. }
    apply (@HAequiv HMK2HA HMK2interp) in D as [D1 D2]. simpl in *.
    destruct (@evalH HMK2HA HMK2interp s); try congruence. exact D2.
  Qed.

End HeytingHMK2.

(** * Independence of Implication and Conjunction *)

Section HeytingHMK3.
  (* McKinsey's Matrix III to prove undefinability of Imp and And *)

  (* Matrix III is actually the direct product of Matrix I with itself *)
  (* Lattice: 1-top 2-a 3-bot
                        11
                       /  \
                      12  21
                     /  \/  \
                    13  22  31
                     \  /\  /
                      23  32
                       \  /
                        33
  *)

  Definition HMK3: Type := prod HMK1 HMK1.

  Definition HMK3R (a b: HMK3) : Prop :=
    match a, b with
      | (a1,a2), (b1, b2) => HMK1R a1 b1 /\ HMK1R a2 b2
    end.

  Definition HMK3meet (a b: HMK3) : HMK3 :=
    match a, b with
      | (a1,a2), (b1, b2) => (HMK1meet a1 b1, HMK1meet a2 b2)
    end.

  Definition HMK3join (a b: HMK3) : HMK3 :=
    match a, b with
      | (a1,a2), (b1, b2) => (HMK1join a1 b1, HMK1join a2 b2)
    end.

  Definition HMK3imp (a b: HMK3) : HMK3 :=
    match a, b with
      | (a1,a2), (b1, b2) => (HMK1imp a1 b1, HMK1imp a2 b2)
    end.

  Lemma HMK3Rref a: HMK3R a a.
  Proof. destruct a as [a b]. destruct a,b; simpl; tauto. Qed.

  Lemma HMK3Rtra a b c: HMK3R a b -> HMK3R b c -> HMK3R a c.
  Proof.
    destruct a as [a1 b1], b as [a2 b2], c as [a3 b3].
    destruct a1,b1,a2,b2,a3,b3; simpl; tauto. Qed.

  Definition HMK3HA : HeytingAlgebra.
  Proof.
  apply (@mkHeytingAlgebra HMK3 HMK3R (HMK1bot, HMK1bot)
                           HMK3meet HMK3join HMK3imp).
  - exact HMK3Rref.
  - exact HMK3Rtra.
  - intros u. destruct u; simpl; tauto.
  - intros [u1 u2] [v1 v2] [w1 w2]. destruct u1, u2, v1, v2, w1, w2; simpl; tauto.
  - intros [u1 u2] [v1 v2] [w1 w2]. destruct u1, u2, v1, v2, w1, w2; simpl; tauto.
  - intros [u1 u2] [v1 v2] [w1 w2]. destruct u1, u2, v1, v2, w1, w2; simpl; tauto.
  Defined.

  (* Idea: Fal, Or, And are closed on {11, 12, 22, 33}, but 11->22 = 21 
     where 1-top 2-a 3-bot *)
  Lemma ImpIndependence_hey:
    ~ exists s, ImpFree s /\ nd [] (Equi (Imp (Var 0) (Var 1)) s).
  Proof.
    intros [s [C D]].
    pose (HMK3interp (z: var) := 
      match z with | 0 => (HMK1top, HMK1a) | _ => (HMK1a, HMK1a) end).
    assert (H: @evalH HMK3HA HMK3interp s el
      [(HMK1top, HMK1top); (HMK1top, HMK1a); (HMK1a, HMK1a); (HMK1bot, HMK1bot)]).
    { clear D. induction s; simpl; auto.
      - destruct v; auto.
      - destruct C as [C1 C2]. specialize (IHs1 C1). specialize (IHs2 C2).
        destruct IHs1 as [H1|[H1|[H1|[H1|H1]]]];
        destruct IHs2 as [H2|[H2|[H2|[H2|H2]]]]; auto;
        rewrite <- H1; rewrite <- H2; auto.
      - destruct C as [C1 C2]. specialize (IHs1 C1). specialize (IHs2 C2).
        destruct IHs1 as [H1|[H1|[H1|[H1|H1]]]];
        destruct IHs2 as [H2|[H2|[H2|[H2|H2]]]]; auto;
        rewrite <- H1; rewrite <- H2; auto. }
    apply (@HAequiv HMK3HA HMK3interp) in D as [D1 D2].
    destruct H as [H|[H|[H|[H|H]]]]; simpl in *; auto; try congruence;
    destruct (evalH (HA:=HMK3HA) HMK3interp s) as [b1 b2];
    destruct b1, b2; try congruence; simpl in D2; tauto.
  Qed.

    (* Idea: Fal, Or, Imp are closed on {11, 12, 13, 31, 33}, but 12/\31 = 32
     where 1-top 2-a 3-bot *)
  Lemma AndIndependence_hey:
    ~ exists s, AndFree s /\ nd [] (Equi (And (Var 0) (Var 1)) s).
  Proof.
    intros [s [C D]].
    pose (HMK3interp (z: var) :=
      match z with | 0 => (HMK1top, HMK1a) | _ => (HMK1bot, HMK1top) end).
    assert (H: @evalH HMK3HA HMK3interp s el
      [(HMK1top, HMK1top); (HMK1top, HMK1a); (HMK1top, HMK1bot); 
       (HMK1bot, HMK1top); (HMK1bot, HMK1bot)]).
    { clear D. induction s; simpl; try tauto.
      - destruct v; auto.
      - destruct C as [C1 C2]. specialize (IHs1 C1). specialize (IHs2 C2).
        destruct IHs1 as [H1|[H1|[H1|[H1|[H1|H1]]]]];
        destruct IHs2 as [H2|[H2|[H2|[H2|[H2|H2]]]]]; try tauto;
        rewrite <- H1; rewrite <- H2; tauto.
      - destruct C as [C1 C2]. specialize (IHs1 C1). specialize (IHs2 C2).
        destruct IHs1 as [H1|[H1|[H1|[H1|[H1|H1]]]]];
        destruct IHs2 as [H2|[H2|[H2|[H2|[H2|H2]]]]]; try tauto;
        rewrite <- H1; rewrite <- H2; tauto. }
    apply (@HAequiv HMK3HA HMK3interp) in D as [D1 D2].
    destruct H as [H|[H|[H|[H|[H|H]]]]]; simpl in *; auto; try congruence;
    destruct (evalH (HA:=HMK3HA) HMK3interp s) as [b1 b2];
    destruct b1, b2; try congruence; simpl in D2; tauto.
  Qed.

End HeytingHMK3.

(** Some underivability results using Heyting *)

Lemma ndXM_hey x:
  ~ nd [] (Or (Var x) (Not (Var x))).
Proof.
  intros D.
  pose (HMK1interp (_ : var) := HMK1a).
  assert (K:= @nd_soundHA HMK1HA HMK1interp [] (Or (Var x) (Not (Var x))) D).
  simpl in K. exact K.
Qed.

Lemma ndDN_hey x:
  ~ nd [] (Imp (Not (Not (Var x))) (Var x)).
Proof.
  intros D.
  pose (HMK1interp (_ : var) := HMK1a).
  assert (K:= @nd_soundHA HMK1HA HMK1interp 
                          [] (Imp (Not (Not (Var x))) (Var x)) D).
  simpl in K. exact K.
Qed.

Lemma ndDN'_hey x:
  ~ nd [] (Or (Not (Var x)) (Not (Not (Var x)))).
Proof.
  intros D.
  pose (HMK2interp (_ : var) := HMK2a).
  assert (K:= @nd_soundHA HMK2HA HMK2interp
                          [] (Or (Not (Var x)) (Not (Not (Var x)))) D).
  simpl in K. exact K.
Qed.

Lemma ndDeMorgan_hey:
  ~ nd [] (Imp (Not (And (Var 0) (Var 1))) (Or (Not (Var 0)) (Not (Var 1)))).
Proof.
  intros D.
  pose (HMK2interp (z : var) := match z with | 0 => HMK2a | _ => HMK2b end).
  assert (K:= @nd_soundHA HMK2HA HMK2interp []
              (Imp (Not (And (Var 0) (Var 1))) (Or (Not (Var 0)) (Not (Var 1))))
              D).
  simpl in K. exact K.
Qed.
